
c***********************************************************************
c
c     The routines pcg.f and pcg1.f implement the Preconditioned 
c     Conjugate Gradient Method (PCG) to solve a symmetric and positive
c     definite system
c   
c                         A*x = b.
c
c     They show how the PREQN package is called from within a
c     Conjugate Gradient code. A user will normally replace these two 
c     routines by his/her own implementation of the PCG method. 
c
c***********************************************************************
c        
      subroutine pcg ( n, nz, a, adiag, jptra, indra, x, b, anorm, 
     *                 w,
     *                 m, iop, iprob, wp, lwp, iwp, liwp, 
     *                 build, info, 
     *                 maxit, tolcg, iout, mssg ) 

      integer  n, nz, jptra(*), indra(*), m, iop, iprob, lwp, iwp(*), 
     *         liwp, info, maxit, iout, mssg
      double   precision a(nz), adiag(n), x(n), b(n), anorm,
     *         w(4*n), wp(*), tolcg
      logical  build
c  
c     This routine partitions the workspace and calls pcg1.f, which
c     implements that PCG method. For a description of the parameters,
c     see routine pcg1.f below.
c
c              * * * * * * * * * * * * * * * * * *
c        
      integer  l1, l2, l3, l4
c
c     Memory partition
c
      l1 = 1
      l2 = l1 + n
      l3 = l2 + n
      l4 = l3 + n
c
      call pcg1 ( n, nz, a, adiag, jptra, indra, x, b, anorm, 
     *            w(l1), w(l2), w(l3), w(l4),
     *            m, iop, iprob, wp, lwp, iwp, liwp, 
     *            build, info, 
     *            maxit, tolcg, iout, mssg )
c      
      return
      end
c
c***********************************************************************
c
      subroutine pcg1   ( n, nz, a, adiag, jptra, indra, x, b, anorm,
     *                    r, v, gv, pr, 
     *                    m, iop, iprob, wp, lwp, iwp, liwp, 
     *                    build, info, 
     *                    maxit, tolcg, iout, mssg ) 
c
      integer  n, nz, jptra(*), indra(*), m, iop, iprob, lwp, iwp(*), 
     *         liwp, info, maxit, iout, mssg
      double   precision a(nz), adiag(n), x(n), b(n), anorm,
     *         r(n), v(n), gv(n), pr(n), wp(*), tolcg
      logical  build

c 
c    This routine implements the PCG method. The preconditioner is
c    computed by the PREQN package
c
c
c     DESCRIPTION OF THE PARAMETERS
c     -----------------------------
c
c     N    is an integer variable specifying the dimension of A
c
c     NZ   is an integer variable specifying the number of nonzeros
c          in A
c     A    is a real array holding the nonzero elements of A in sparse
c          column oriented format. Given that A is symmetric, only its 
c          lower triangle is stored.
c
c     ADIAG  is a real array of dimension NZ containing the diagonal 
c          elements of matrix A.
c
c     JPTRA  integer array of dimension N+1 holding pointers to the 
c          columns of A in the column oriented format.
c
c     INDRA  integer array of dimension NZ containing row indices of 
c          the elements in A.
c
c     X    real array of dimension N. On input it contains an initial 
c          approximation. On output it contains the last approximation 
c          found by the CG method.
c
c     B    real array of dimension N holding the elements of the 
c          right hand side
c
c     ANORM  real variable holding the infinity norm of the matrix A. 
c          This quantity is used in the termination test.
c
c     W    double precision array of length 4N. It is used as 
c          real work space to hold intermediate computations.
c
c     M    is an integer variable specifying the maximum number of pairs
c          used in the quasi-Newton preconditioner.
c
c     IOP  is an integer variable specifying the storing scheme of the 
c          pairs {sk,yk}.
c
c          IOP = 1  the pairs are sampled in order to have
c                   an almost uniform distribution.
c
c          IOP = 2  the last M consecutive pairs are stored.
c
c     IPROB  is an integer labeling the linear system being solved 
c
c     WP   is a real array of dimension no less than 
c          4(M+1)N + 2(M+1) + 1.  It is used as real workspace to
c          compute the quasi-Newton preconditioner.
c
c     LWP  is an integer variable specifying length of the array W as 
c          declared in the calling program.
c
c     IWP  is an integer array of dimension no less than 2M + 30. It 
c          is used as integer workspace to compute the quasi-Newton 
c          preconditioner.
c
c     LIWP is an integer specifying the length of the array IW as 
c          declared in the calling subprogram.
c
c     BUILD is a logical variable that determines if the preconditioner
c          stored while solving IPROB-1 should be built. This variable
c          allows the reuse of a preconditioner.
c
c     INFO  is an integer variable. On output it gives information
c          about the status of the routine.
c
c          INFO =  0   normal termination
c          INFO = -1   error in input parameters
c          INFO = -2   insufficient storage in the workspace arrays
c          INFO = -4   initial residual is sufficiently small
c         
c    MAXIT integer variable specifying the maximum number of CG 
c          iterations
c
c    TOLCG Tolerance to stop the CG iteration.
c
c    IOUT  is an integer variable that controls the amount and type of
c          output
c
c          IOUT = 0   no output printed
c          IOUT = >0  output at every IOUT iterations of the CG method, 
c                    plus output at first and last iteration.
c
c    MSSG  is an integer variable specifying the output unit used 
c          to display error messages
c 
c              * * * * * * * * * * * * * * * * * *

c
      integer  jcg, i
      double   precision dnrmif, bnorm, xnorm, rnorm, resrel, 
     *                   rho, ddot, beta, rhold, vgv, alpha
c
c     Check input 
c
      info =  0

      if ( n.le.0 .or. m.lt.0 .or. maxit .le.0 ) then 
         write(mssg,900)
         info = -1
         return
      end if
c
c     Initial printing
c
      if ( iout.ne.0 ) write(6,920) m, iop, iprob
c     
c     Compute the initial residual. First compute the sparse product
c     Ax = v using the routine dsymvs from the MINPACK-2 package.
c
      call dsymvs ( n, a, adiag, jptra, indra, x, v )
      do 100 i = 1, n
         r(i)  = b(i) - v(i)
 100  continue
c
c     if the initial residual is sufficiently small, terminate
c
      bnorm = dnrmif ( n, b )
      rnorm = dnrmif ( n, r )
      if ( rnorm.le.bnorm*tolcg ) then
         info = -4
         return
      end if
c
c*************************
c     Begin CG iteration
c*************************
c
      do 300 jcg = 1, maxit
c
c        Compute the preconditioned residual using the PREQN package
c
         call precnd ( n, m, iop, iprob, jcg, v, gv, r, pr, wp, lwp,
     *                 iwp, liwp, build, info, mssg )
         if ( info.ne.0 ) then
            write(mssg,940)
c
            return
         end if 
c--------
c        Check convergence. Print the relative residual at the initial
c        and final iterations, as well as at selected iterations.
c        The stopping test is  
c
c               ||r|| < tolcg*( ||A|| ||x|| + ||b|| )
c
c        where ||*|| denotes the infinity norm.         
c
         xnorm  = dnrmif ( n, x )
         rnorm  = dnrmif ( n, r )
         resrel = rnorm /( anorm*xnorm + bnorm )
c
         if ( iout.ne.0 ) then
            if ( jcg.eq.1 ) then 
               write(6,960) jcg-1, resrel
            else
               if ( mod(jcg-1,iout).eq.0 ) write(6,960) jcg-1, resrel
            end if
         end if
c
         if ( resrel.le.tolcg ) then 
            if ( iout.ne.0 ) write(6,960) jcg-1, resrel
            return
         end if
c--------
c
c        Compute a new direction v       
c
         rho = ddot ( n, r, 1, pr, 1 )
c
         if ( jcg.eq.1 ) then
            call dcopy ( n, pr, 1, v, 1 )
         else
            beta = rho/rhold
c
            do 200 i=1, n
               v(i) = pr(i) + beta*v(i)
 200        continue
         end if
c
c     Compute the matrix-vector product Av = gv
c
         call dsymvs ( n, a, adiag, jptra, indra, v, gv )
c
         vgv = ddot ( n, v, 1, gv, 1 )
c
c     Compute step length, new iterate and new residual
c
         alpha = rho / vgv
         call daxpy ( n,  alpha,  v, 1, x, 1 )
         call daxpy ( n, -alpha, gv, 1, r, 1 )        
c
         rhold = rho
 300  continue
c
c**************************
c     end of CG iteration
c**************************
c
 900     format(//,2x,' * Error in PCG: incorrect input parameters',/)
 920     format(////,15x,'*** New Problem ***',/, 2x,
     *           'Preconditioned Conjugate Gradient method using ',/,
     *       2x, 'a quasi-Newton preconditioner ', //,
     *       5x, 'number of pairs in the preconditioner = ', i3,/,
     *       5x, 'sampling strategy                     = ', i3,/,
     *       5x, 'number of problem                     = ', i3,//,
     *       5x,'||rel|| is the norm of the relative residual',/)
 940     format(//,2x,' * Abnormal termination in PCG.', 
     *                ' Error in PREQN',/)
 960     format( 5x,' iter = ',i6, 2x,' ||rel|| = ',
     *           e14.8)

      return
      end
c
c***********************************************************************
      
     





















